''' Depth to PointCloud Tensorflow Subgraph
'''
from __future__ import print_function

import sys
import os
import tensorflow as tf
import numpy as np
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(BASE_DIR)
sys.path.append(BASE_DIR)
sys.path.append(os.path.join(ROOT_DIR, 'utils'))
#import tf_util
sys.path.append(os.path.join(ROOT_DIR, 'train'))
import provider_e2e
from train_util_e2e import get_batch

from data_prep_e2e_tf import depthmap_to_frustum_pointcloud
from data_prep_e2e_tf_rot import depthmap_to_frustum_pointcloud_rot

ROTATE_TO_CENTER = True

pickle_train_file = "/home/EE367_project/frustum-pointnets/kitti/frustum_preprocessed_data_training.pkl"

TRAIN_DATASET = provider_e2e.FrustumDataset(split='training', pkl_file_path = pickle_train_file,
    rotate_to_center=ROTATE_TO_CENTER, random_flip=False, random_shift=False, one_hot=True)


def train():
    ''' Main function for training and simple evaluation. '''
    with tf.Graph().as_default():
        BATCH_SIZE = 32
        npy_depth = tf.placeholder(tf.float32, shape = (BATCH_SIZE, 375,1242))
        calib_mat = tf.placeholder(tf.float32, shape = (BATCH_SIZE, 3, 4)) 
        box2d = tf.placeholder(tf.float32, shape = (BATCH_SIZE, 4)) 
        box3d = tf.placeholder(tf.float32, shape = (BATCH_SIZE, 8, 3)) 
        rot_angle = tf.placeholder(tf.float32, shape = (BATCH_SIZE, ))
        
        sess = tf.Session()
        
        if ROTATE_TO_CENTER:
            pc_rect_result = depthmap_to_frustum_pointcloud_rot(npy_depth, box2d, box3d, calib_mat, rot_angle)
        else: 
            pc_rect_result = depthmap_to_frustum_pointcloud(npy_depth, box2d, box3d, calib_mat)
                
        
        ops = {'npy_depth_pl': npy_depth, 
               'calib_mat_pl': calib_mat, 
               'box2d_pl': box2d,
               'box3d_pl': box3d,
               'pc_rect_result': pc_rect_result,
               'rot_angle_pl': rot_angle}
        
        train_idxs = np.arange(0, len(TRAIN_DATASET))
        
        for batch_idx in range(1):
            start_idx = batch_idx * BATCH_SIZE
            end_idx = (batch_idx+1) * BATCH_SIZE
            
            npy_arrays, imgs, data_idx, calibs, box2d, box3d, batch_center, \
            batch_hclass, batch_hres, \
            batch_sclass, batch_sres, \
            batch_rot_angle, batch_one_hot_vec = \
                get_batch(TRAIN_DATASET, train_idxs, start_idx, end_idx, 3)

            feed_dict = {ops['npy_depth_pl']: npy_arrays, 
                         ops['calib_mat_pl']: calibs, 
                         ops['box2d_pl']: box2d, 
                         ops['box3d_pl']: box3d, 
                         ops['rot_angle_pl']: batch_rot_angle}

            result, labels = sess.run(ops['pc_rect_result'], feed_dict=feed_dict)
            
            print(result.shape)
            print(labels.shape)
            np.save('testrun_outputs/testrun_pc.npy', result)
            np.save('testrun_outputs/box2d.npy', box2d)
            with open("testrun_outputs/testrun_ids.txt", "wb") as f:
                np.savetxt(f, data_idx, fmt='%d', delimiter="\n")


if __name__ == "__main__":
    train()