''' Util functions for training and evaluation.

Author: Charles R. Qi
Date: September 2017
'''

import numpy as np
import pdb

def get_batch(dataset, idxs, start_idx, end_idx, num_channel,
              from_rgb_detection=False):
    ''' Prepare batch data for training/evaluation.
    batch size is determined by start_idx-end_idx

    Input:
        dataset: an instance of FrustumDataset class
        idxs: a list of data element indices
        start_idx: int scalar, start position in idxs
        end_idx: int scalar, end position in idxs
        num_point: int scalar
        num_channel: int scalar
        from_rgb_detection: bool
    Output:
        batched data and label
    '''
    if from_rgb_detection:
        return get_batch_from_rgb_detection(dataset, idxs, start_idx, end_idx,
            num_point, num_channel)

    bsize = end_idx-start_idx
    #batch_data = np.zeros((bsize, num_point, num_channel))
    #batch_label = np.zeros((bsize, num_point), dtype=np.int32)
    batch_npy = np.zeros((bsize, 375, 1242))
    batch_img = np.zeros((bsize, 375, 1242, 3))
    data_idx_arr = np.zeros((bsize))
    calib_P_matrix = np.zeros((bsize, 3, 4))
    box2d = np.zeros((bsize, 4))
    box3d = np.zeros((bsize, 8, 3))
    batch_center = np.zeros((bsize, 3))
    batch_heading_class = np.zeros((bsize,), dtype=np.int32)
    batch_heading_residual = np.zeros((bsize,))
    batch_size_class = np.zeros((bsize,), dtype=np.int32)
    batch_size_residual = np.zeros((bsize, 3))
    batch_rot_angle = np.zeros((bsize,))
    if dataset.one_hot:
        batch_one_hot_vec = np.zeros((bsize,3)) # for car,ped,cyc
    for i in range(bsize):
        if dataset.one_hot:
            npy, img , data_idx ,calib, bbox2d, bbox3d, center,hclass,hres,sclass,sres,rotangle,onehotvec = \
                dataset[idxs[i+start_idx]]
            batch_one_hot_vec[i] = onehotvec
        else:
            npy, img, data_idx, calib, bbox2d, bbox3d, center,hclass,hres,sclass,sres,rotangle = \
                dataset[idxs[i+start_idx]]
        
        batch_npy[i,...] = npy
        batch_img[i, ...] = img
        data_idx_arr[i] = data_idx
        calib_P_matrix[i,...] = calib

        box2d[i,...] = bbox2d
        box3d[i,...] = bbox3d
        
        #batch_data[i,...] = ps[:,0:num_channel]
        #batch_label[i,:] = seg
        batch_center[i,:] = center
        batch_heading_class[i] = hclass
        batch_heading_residual[i] = hres
        batch_size_class[i] = sclass
        batch_size_residual[i] = sres
        batch_rot_angle[i] = rotangle
        
    if dataset.one_hot:
        return batch_npy, batch_img, data_idx_arr,calib_P_matrix, box2d, box3d, batch_center, \
            batch_heading_class, batch_heading_residual, \
            batch_size_class, batch_size_residual, \
            batch_rot_angle, batch_one_hot_vec
    else:
        return batch_npy, batch_img, data_idx_arr, calib_P_matrix, box2d, box3d, batch_center, \
            batch_heading_class, batch_heading_residual, \
            batch_size_class, batch_size_residual, batch_rot_angle

def get_batch_from_rgb_detection(dataset, idxs, start_idx, end_idx,
                                 num_point, num_channel):
    bsize = end_idx-start_idx
    batch_data = np.zeros((bsize, num_point, num_channel))
    batch_rot_angle = np.zeros((bsize,))
    batch_prob = np.zeros((bsize,))
    if dataset.one_hot:
        batch_one_hot_vec = np.zeros((bsize,3)) # for car,ped,cyc
    for i in range(bsize):
        if dataset.one_hot:
            ps,rotangle,prob,onehotvec = dataset[idxs[i+start_idx]]
            batch_one_hot_vec[i] = onehotvec
        else:
            ps,rotangle,prob = dataset[idxs[i+start_idx]]
        batch_data[i,...] = ps[:,0:num_channel]
        batch_rot_angle[i] = rotangle
        batch_prob[i] = prob
    if dataset.one_hot:
        return batch_data, batch_rot_angle, batch_prob, batch_one_hot_vec
    else:
        return batch_data, batch_rot_angle, batch_prob


