import numpy as np
import cv2
import glob
import os

def get_npy_filenames(folder_path):    
    return glob.glob(os.path.join(folder_path, '*.npy'))

input_folder = "depth_files_training/"
output_folder = "real_depth_training/"
depth_files = sorted(get_npy_filenames(input_folder))

# conversion from pixels to real world using focal length in pixels
width_to_focal = dict()
width_to_focal[1242] = 721.5377

h_pre = 311
w_pre = 1242

h_full = 375
w_full = 1242

num_files = len(depth_files)

for index, dp_file in enumerate(depth_files):
    print(index, " out of ", num_files) 
    
    # 0. load disparity map (prediction)
    dp = np.load(dp_file)

    # 1. rescale to kitti size, and add NAN (-1) values at the top again
    dp_r = cv2.resize(dp, (w_pre, h_pre), interpolation=cv2.INTER_LINEAR)
    dp_r = np.concatenate([-1*np.ones((h_full-h_pre, w_full)), dp_r], axis=0)
    dp_r[dp_r<=0] = -1.0
    
    # 1.1 adjust scaling, so that bottom row of pixels is at the same disparity, always!
    factor = 169.66 # estimated scaling factor, dows not include / focal_len_pix
    crop = dp_r[-10:, 400:600]
    sumVal = np.sum(crop)
    dp_r = dp_r / sumVal * factor

    # 2. convert to real world using the following formula
    real_depth = width_to_focal[w_full] * 0.54 / dp_r
    
    # 3. save real depth to output folder
    idx = dp_file.rfind('/') + 1
    file_name = output_folder + "/" + dp_file[idx:-3] + "npy"
    np.save(file_name, real_depth)