import tensorflow as tf
import numpy as np

# conversion from pixels to real world using focal length in pixels
focal_length_pixels = 721.5377

h_pre = 311
w_pre = 1242

h_full = 375
w_full = 1242

BATCH_SIZE = 16

def convert_disp_to_depth(disp):
    
    BATCH_SIZE = tf.shape(disp)[0]

    # 1. rescale to kitti size, and add NAN (-1) values at the top again
    disp_r = tf.image.resize_bilinear(disp, [h_pre, w_pre])
    add_top =  -1*tf.ones([BATCH_SIZE, h_full-h_pre, w_full, 1])
    disp_r = tf.concat([add_top, disp_r], axis = 1, name = 'add_img_top')
    disp_r = tf.squeeze(disp_r) # (B, h, w)
    
    # 1.1 adjust scaling, so that bottom row of pixels is at the same disparity, always!
    factor = 169.66 # estimated scaling factor, dows not include / focal_len_pix
    
    def enforce_scale(disp_r):
        crop = disp_r[:, -10:, 400:600] # (B, 10, 200)
        sumVal = tf.reduce_sum(crop, axis=[1, 2]) # (B,)
        disp_r = tf.transpose(tf.divide(tf.transpose(disp_r), sumVal)) * factor        
        return disp_r
    
    disp_r = enforce_scale(disp_r)
    disp_r = tf.reshape(disp_r, [BATCH_SIZE, h_full, w_full])

    # 2. convert to real world using the following formula
    real_depth = tf.divide(focal_length_pixels * 0.54, disp_r)

    return real_depth