# Copyright UCL Business plc 2017. Patent Pending. All rights reserved.
#
# The MonoDepth Software is licensed under the terms of the UCLB ACP-A licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.
#
# For any other use of the software not covered by the UCLB ACP-A Licence,
# please contact info@uclb.com

from __future__ import absolute_import, division, print_function

# only keep warnings and errors
import os
os.environ['TF_CPP_MIN_LOG_LEVEL']='0'

import glob
import sys

import numpy as np
import argparse
import re
import time
import tensorflow as tf
import tensorflow.contrib.slim as slim
import scipy.misc
import matplotlib.pyplot as plt

from monodepth_model import *
from monodepth_dataloader import *
from average_gradients import *

from PIL import Image

import pdb
#parser = argparse.ArgumentParser(description='Monodepth TensorFlow implementation.')

#parser.add_argument('--encoder',          type=str,   help='type of encoder, vgg or resnet50', default='vgg')
#parser.add_argument('--image_path',       type=str,   help='path to the image', required=True)
#parser.add_argument('--checkpoint_path',  type=str,   help='path to a specific checkpoint to load', required=True)
#parser.add_argument('--input_height',     type=int,   help='input height', default=256)
#parser.add_argument('--input_width',      type=int,   help='input width', default=512)

#args = parser.parse_args()

def get_image_filenames(folder_path):    
    return glob.glob(os.path.join(folder_path, '*.png'))

def post_process_disparity(disp):
    _, h, w = disp.shape
    l_disp = disp[0,:,:]
    r_disp = np.fliplr(disp[1,:,:])
    m_disp = 0.5 * (l_disp + r_disp)
    l, _ = np.meshgrid(np.linspace(0, 1, w), np.linspace(0, 1, h))
    l_mask = 1.0 - np.clip(20 * (l - 0.05), 0, 1)
    r_mask = np.fliplr(l_mask)
    return r_mask * l_disp + l_mask * r_disp + (1.0 - l_mask - r_mask) * m_disp

def test_simple(params):
    """Test function."""

    image_folder = '/home/EE367_project/frustum-pointnets/dataset/KITTI/object/training/image_2'
    image_files = sorted(get_image_filenames(image_folder))
    
    input_height = 256
    input_width = 1024

    original_height = 375
    original_width = 1242

    left  = tf.placeholder(tf.float32, [2, input_height, input_width, 3])
    print("Got here 1")
    model = MonodepthModel(params, "test", left, None)
    print("Got here 2")
    #input_image = scipy.misc.imread(args.image_path, mode="RGB")
    #original_height, original_width, num_channels = input_image.shape
    #input_image = scipy.misc.imresize(input_image, [args.input_height, args.input_width], interp='lanczos')
    #input_image = input_image.astype(np.float32) / 255
    #input_images = np.stack((input_image, np.fliplr(input_image)), 0)

    # SESSION
    config = tf.ConfigProto(allow_soft_placement=True)
    sess = tf.Session(config=config)

    # SAVER
    train_saver = tf.train.Saver()

    # INIT
    sess.run(tf.global_variables_initializer())
    sess.run(tf.local_variables_initializer())
    coordinator = tf.train.Coordinator()
    threads = tf.train.start_queue_runners(sess=sess, coord=coordinator)

    # RESTORE
    restore_path = "./model_kitti" #args.checkpoint_path.split(".")[0]
    train_saver.restore(sess, restore_path)
    print("Model restored")
    num_imgs = len(image_files)
    #input_height = 375
    #inout_width = 1242
    for index, img_path in enumerate(image_files):
        print(index, " out of ", num_imgs) 
        img = Image.open(img_path)
        img = img.resize([input_width, input_height], Image.ANTIALIAS)
        input_image = np.array(img).astype(np.float32) / 255.0
        input_images = np.stack((input_image, np.fliplr(input_image)), 0)
        #input_image = scipy.misc.imresize(input_image, [args.input_height, args.input_width], interp='lanczos')
        disp = sess.run(model.disp_left_est[0], feed_dict={left: input_images})
        disp_pp = post_process_disparity(disp.squeeze()).astype(np.float32)
        disp_to_img = scipy.misc.imresize(disp_pp.squeeze(), [original_height, original_width])

        #output_directory = "./depth_files/"
        idx = img_path.rfind('/') + 1
        file_name = "./depth_files/" + img_path[idx:-3] + "npy"
        #output_name = os.path.splitext(os.path.basename(args.image_path))[0]
        np.save(file_name, disp_pp)
        #np.save(os.path.join(output_directory, "{}_disp.npy".format(output_name)), disp_pp)
        #disp_to_img = scipy.misc.imresize(disp_pp.squeeze(), [input_height, input_width])
        #plt.imsave(os.path.join(output_directory, "{}_disp.png".format(output_name)), disp_to_img, cmap='plasma')

    print('done!')

def main():
    
    params = monodepth_parameters(
        encoder='vgg',
        height=375,
        width=1242,
        batch_size=2,
        num_threads=1,
        num_epochs=1,
        do_stereo=False,
        wrap_mode="border",
        use_deconv=False,
        alpha_image_loss=0,
        disp_gradient_loss_weight=0,
        lr_loss_weight=0,
        full_summary=False)

    test_simple(params)

if __name__ == '__main__':
    main()
    #tf.app.run()
