import numpy as np
import tensorflow as tf
import cv2


focal_length = 721.5377

h_in = 256
w_in = 1024

h_pre = 311
w_pre = 1242

h_full = 375
w_full = 1242

def disp_to_depth_np(npy_file):
    
    real_depth = np.zeros((npy_file.shape[0], h_full, w_full))
    
    return_stuff = np.zeros((npy_file.shape[0], h_pre, w_pre,1))
    
    for i in range(npy_file.shape[0]):
        npy = np.squeeze(npy_file[i])

        # 1. rescale to kitti size, and add NAN (-1) values at the top again
        dp_r = cv2.resize(npy, (w_pre, h_pre), interpolation=cv2.INTER_AREA)
        dp_c = np.concatenate([-1*np.ones((h_full-h_pre, w_full)), dp_r], axis=0)
        #dp_r[dp_r<=0] = -1.0

        # 1.1 adjust scaling, so that bottom row of pixels is at the same disparity, always!
        factor = 169.66 # estimated scaling factor, dows not include / focal_len_pix
        crop = dp_c[-10:, 400:600]
        sumVal = np.sum(crop)
        dp_t = dp_c / sumVal * factor

        # 2. convert to real world using the following formula
        real_depth[i, ...] = focal_length * 0.54 / dp_t
        return_stuff[i,:,:, 0] = dp_r
    
    return np.squeeze(return_stuff)


def disp_to_depth_tf(disp):
    
    BATCH_SIZE = tf.shape(disp)[0]

    # 1. rescale to kitti size, and add NAN (-1) values at the top again
    disp_r = tf.image.resize_area(disp, [h_pre, w_pre])
    add_top =  -1*tf.ones([BATCH_SIZE, h_full-h_pre, w_full, 1])
    disp_c = tf.concat([add_top, disp_r], axis = 1, name = 'add_img_top')
    disp_c = tf.squeeze(disp_c) # (B, h, w)
    
    # 1.1 adjust scaling, so that bottom row of pixels is at the same disparity, always!
    factor = 169.66 # estimated scaling factor, dows not include / focal_len_pix
    
    def enforce_scale(disp):
        crop = disp[:, -10:, 400:600] # (B, 10, 200)
        sumVal = tf.reduce_sum(crop, axis=[1, 2]) # (B,)
        disp = tf.transpose(tf.divide(tf.transpose(disp), sumVal)) * factor        
        return disp
    
    disp_t = enforce_scale(disp_c)
    disp_t = tf.reshape(disp_t, [BATCH_SIZE, h_full, w_full])

    # 2. convert to real world using the following formula
    real_depth = tf.divide(focal_length * 0.54, disp_t)

    return tf.squeeze(disp_r)


np.random.seed(0)
a = 100
b = 400
test_value_real = np.zeros((2, h_pre, w_pre, 1))
test_value = (b - a) * (np.random.ranf(size = (2, h_in, w_in, 1))) + a
for i in range(test_value.shape[0]):
    npy = np.squeeze(test_value[i])

    # 1. rescale to kitti size, and add NAN (-1) values at the top again
    dp_r = cv2.resize(npy, (w_pre, h_pre), interpolation=cv2.INTER_AREA)
    test_value_real[i,:, :, 0] = dp_r

res_np = disp_to_depth_np(test_value)

with tf.Graph().as_default():
        npy_depth = tf.placeholder(tf.float32, shape = (2, h_in, w_in, 1))
        out = disp_to_depth_tf(npy_depth)
        
        with tf.Session() as sess:
            res_tf = sess.run(out, feed_dict={npy_depth: test_value})
            
diff = np.abs(res_tf - res_np)
print(np.mean(diff))

