# Copyright UCL Business plc 2017. Patent Pending. All rights reserved.
#
# The MonoDepth Software is licensed under the terms of the UCLB ACP-A licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.
#
# For any other use of the software not covered by the UCLB ACP-A Licence,
# please contact info@uclb.com

from __future__ import absolute_import, division, print_function

# only keep warnings and errors
import os
os.environ['TF_CPP_MIN_LOG_LEVEL']='1'
os.environ["CUDA_VISIBLE_DEVICES"]= "0, 1"

import sys

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(BASE_DIR)
sys.path.append(BASE_DIR)
sys.path.append("/home/EE367_project/frustum-pointnets/models")
sys.path.append("/home/EE367_project/frustum-pointnets/kitti/")
sys.path.append("/home/EE367_project/frustum-pointnets/train/")
sys.path.append("/home/EE367_project/monodepth/")

import numpy as np
import argparse
import re
import time
import tensorflow as tf
import tensorflow.contrib.slim as slim

from monodepth_model_e2e import *
from monodepth_dataloader_e2e import *

import glob
import importlib
import pandas as pd 
from datetime import datetime
import provider_e2e
from train_util_e2e import get_batch
import time
import pdb

from data_prep_e2e_tf import depthmap_to_frustum_pointcloud
from data_prep_e2e_tf_rot import depthmap_to_frustum_pointcloud_rot

parser = argparse.ArgumentParser(description='Monodepth TensorFlow implementation.')

parser.add_argument('--mode',                      type=str,   help='train or test', default='test')
parser.add_argument('--model_name',                type=str,   help='model name', default='monodepth_inference_e2e')
parser.add_argument('--encoder',                   type=str,   help='type of encoder, vgg or resnet50', default='vgg')
parser.add_argument('--dataset',                   type=str,   help='dataset to train on, kitti, or cityscapes', default='kitti')
#parser.add_argument('--data_path',                 type=str,   help='path to the data', required=True)
#parser.add_argument('--filenames_file',            type=str,   help='path to the filenames text file', required=True)
#parser.add_argument('--input_height',              type=int,   help='input height', default=256)
#parser.add_argument('--input_width',               type=int,   help='input width', default=512)
parser.add_argument('--batch_size',                type=int,   help='batch size', default=16)
parser.add_argument('--num_epochs',                type=int,   help='number of epochs', default=50)
parser.add_argument('--learning_rate',             type=float, help='initial learning rate', default=1e-4)
parser.add_argument('--lr_loss_weight',            type=float, help='left-right consistency weight', default=1.0)
parser.add_argument('--alpha_image_loss',          type=float, help='weight between SSIM and L1 in the image loss', default=0.85)
parser.add_argument('--disp_gradient_loss_weight', type=float, help='disparity smoothness weigth', default=0.1)
parser.add_argument('--do_stereo',                             help='if set, will train the stereo model', action='store_true')
parser.add_argument('--wrap_mode',                 type=str,   help='bilinear sampler wrap mode, edge or border', default='border')
parser.add_argument('--use_deconv',                            help='if set, will use transposed convolutions', action='store_true')
parser.add_argument('--num_gpus',                  type=int,   help='number of GPUs to use for training', default=1)
parser.add_argument('--num_threads',               type=int,   help='number of threads to use for data loading', default=8)
parser.add_argument('--output_directory',          type=str,   help='output directory for test disparities, if empty outputs to checkpoint folder', default='')
#parser.add_argument('--log_directory',             type=str,   help='directory to save checkpoints and summaries', default='')
#parser.add_argument('--checkpoint_path',           type=str,   help='path to a specific checkpoint to load', default='')
parser.add_argument('--retrain',                               help='if used with checkpoint_path, will restart training from step zero', action='store_true')
parser.add_argument('--full_summary',                          help='if set, will keep more data for each summary. Warning: the file can become very large', action='store_true')

args = parser.parse_args()

# Monodepth parameters
DATA_PATH = "/home/EE367_project/frustum-pointnets/dataset/KITTI"
FILENAMES_FILE = "monodepth/image_sets/train.txt"
CHECKPOINT_PATH = "/home/EE367_project/monodepth/log/model_kitti_test/model-23200"
LOG_DIR = "monodepth/log_e2e/"
INPUT_HEIGHT = 256
INPUT_WIDTH = 1024

# Frustum Parameters
FRUSTUM_MODEL = "frustum_pointnets_v1_e2e"
EPOCH_CNT = 0
BATCH_SIZE = args.batch_size
NUM_POINT = 1024
MAX_EPOCH = 50
BASE_LEARNING_RATE = 0.001
GPU_INDEX = 0
MOMENTUM = 0.9
OPTIMIZER = 'adam'
DECAY_STEP = 200000
DECAY_RATE = 0.7
NUM_CHANNEL = 3
NUM_CLASSES = 2 # segmentation has two classes
RESTORE_MODEL_PATH = "/home/EE367_project/frustum-pointnets/snapshots/log_v1/model.ckpt"

MODEL = importlib.import_module(FRUSTUM_MODEL) # import network module
MODEL_FILE = os.path.join(ROOT_DIR, 'models', FRUSTUM_MODEL+'.py')
if not os.path.exists(LOG_DIR): os.mkdir(LOG_DIR)
os.system('cp %s %s' % (MODEL_FILE, LOG_DIR)) # bkp of model def
os.system('cp %s %s' % (os.path.join(BASE_DIR, 'train.py'), LOG_DIR))
#LOG_FOUT = open(os.path.join(LOG_DIR, 'log_train.txt'), 'w')
#LOG_FOUT.write(str(FLAGS)+'\n')

BN_INIT_DECAY = 0.5
BN_DECAY_DECAY_RATE = 0.5
BN_DECAY_DECAY_STEP = float(DECAY_STEP)
BN_DECAY_CLIP = 0.99

# Load Frustum Datasets. Use default data paths.

pickle_train_file = "/home/EE367_project/frustum-pointnets/kitti/frustum_preprocessed_data_training.pkl"
pickle_val_file = "/home/EE367_project/frustum-pointnets/kitti/pickles/online_trainin_data_validation.pkl"

ROTATE_TO_CENTER = True


def log_string(out_str):
    LOG_FOUT.write(out_str+'\n')
    LOG_FOUT.flush()
    print(out_str)

def get_learning_rate(batch):
    learning_rate = tf.train.exponential_decay(
                        BASE_LEARNING_RATE,  # Base learning rate.
                        batch * BATCH_SIZE,  # Current index into the dataset.
                        DECAY_STEP,          # Decay step.
                        DECAY_RATE,          # Decay rate.
                        staircase=True)
    learing_rate = tf.maximum(learning_rate, 0.00001) # CLIP THE LEARNING RATE!
    return learning_rate        

def get_bn_decay(batch):
    bn_momentum = tf.train.exponential_decay(
                      BN_INIT_DECAY,
                      batch*BATCH_SIZE,
                      BN_DECAY_DECAY_STEP,
                      BN_DECAY_DECAY_RATE,
                      staircase=True)
    bn_decay = tf.minimum(BN_DECAY_CLIP, 1 - bn_momentum)
    return bn_decay

def post_process_disparity(disp):
    _, h, w = disp.shape
    l_disp = disp[0,:,:]
    r_disp = np.fliplr(disp[1,:,:])
    m_disp = 0.5 * (l_disp + r_disp)
    l, _ = np.meshgrid(np.linspace(0, 1, w), np.linspace(0, 1, h))
    l_mask = 1.0 - np.clip(20 * (l - 0.05), 0, 1)
    r_mask = np.fliplr(l_mask)
    return r_mask * l_disp + l_mask * r_disp + (1.0 - l_mask - r_mask) * m_disp

def count_text_lines(file_path):
    f = open(file_path, 'r')
    lines = f.readlines()
    f.close()
    return len(lines)

def both_models_e2e(params):
    """Test function."""
      
    # load and restore the monodepth graph
    tf_monodepth_graph = tf.Graph()
    with tf_monodepth_graph.as_default():

        dataloader = MonodepthDataloader(DATA_PATH, FILENAMES_FILE, params, args.dataset, args.mode)
        left  = dataloader.left_image_batch

        model = MonodepthModel(params, args.mode, left, None)

        # worth a try
        model.real_depth_est = tf.identity(model.real_depth_est, name = "real_depth")

        # SESSION
        config = tf.ConfigProto(allow_soft_placement=True)
        sess = tf.Session(config=config)

        # SAVER
        #train_saver = tf.train.Saver()

        # INIT
        sess.run(tf.global_variables_initializer())
        sess.run(tf.local_variables_initializer())
        coordinator = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coordinator)

#         # RESTORE
#         if CHECKPOINT_PATH == '':
#             restore_path = tf.train.latest_checkpoint(args.log_directory + '/' + args.model_name)
#         else:
#             restore_path = CHECKPOINT_PATH.split(".")[0]
#         train_saver.restore(sess, restore_path)

        real_depth = tf_monodepth_graph.get_tensor_by_name("real_depth:0")
        real_depth = tf.reshape(real_depth, [BATCH_SIZE, 375, 1242])
        
        print("Monodepth model sucessfully restored!")

        # load and restore the frustum graph
    #     tf_frustum_graph = tf.Graph()
    #     with tf_frustum_graph.as_default():

        npy_depth_pl, rgb_img_pl, box2d_pl, box3d_pl, rot_angle_pl, calib_pl, one_hot_vec_pl, centers_pl, \
        heading_class_label_pl, heading_residual_label_pl, \
        size_class_label_pl, size_residual_label_pl = \
            MODEL.placeholder_inputs(BATCH_SIZE, NUM_POINT)

        is_training_pl = tf.placeholder(tf.bool, shape = ())

        # Note the global_step=batch parameter to minimize. 
        # That tells the optimizer to increment the 'batch' parameter
        # for you every time it trains.
        batch = tf.get_variable('batch', [],
            initializer=tf.constant_initializer(0), trainable=False)
        bn_decay = get_bn_decay(batch)
        tf.summary.scalar('bn_decay', bn_decay)

        # Get model and losses (note: no placehoder anymore as first argument!)
        if ROTATE_TO_CENTER:
            pointcloud, labels = \
                depthmap_to_frustum_pointcloud_rot(real_depth, box2d_pl, box3d_pl, \
                                                   calib_pl, rot_angle_pl, npoints = NUM_POINT)
        else:
            pointcloud, labels = \
                depthmap_to_frustum_pointcloud(real_depth, box2d_pl, box3d_pl, \
                                               calib_pl, npoints = NUM_POINT)
        pointcloud = tf.reshape(pointcloud, [BATCH_SIZE, NUM_POINT, 4])

        end_points = MODEL.get_model(pointcloud, one_hot_vec_pl,
            is_training_pl, bn_decay=bn_decay)
        loss = MODEL.get_loss(labels, centers_pl,
            heading_class_label_pl, heading_residual_label_pl,
            size_class_label_pl, size_residual_label_pl, end_points)
        tf.summary.scalar('loss', loss)

        variables = tf.trainable_variables() 
        lossL2 = tf.add_n([ tf.nn.l2_loss(v) for v in variables ]) * 0.001
        loss += lossL2

        losses = tf.get_collection('losses')
        total_loss = tf.add_n(losses, name='total_loss')
        tf.summary.scalar('total_loss', total_loss)

        # Write summaries of bounding box IoU and segmentation accuracies
        iou2ds, iou3ds = tf.py_func(provider_e2e.compute_box3d_iou, [\
            end_points['center'], \
            end_points['heading_scores'], end_points['heading_residuals'], \
            end_points['size_scores'], end_points['size_residuals'], \
            centers_pl, \
            heading_class_label_pl, heading_residual_label_pl, \
            size_class_label_pl, size_residual_label_pl], \
            [tf.float32, tf.float32])
        end_points['iou2ds'] = iou2ds 
        end_points['iou3ds'] = iou3ds 
        tf.summary.scalar('iou_2d', tf.reduce_mean(iou2ds))
        tf.summary.scalar('iou_3d', tf.reduce_mean(iou3ds))

        correct = tf.equal(tf.argmax(end_points['mask_logits'], 2),
            tf.to_int64(labels))
        accuracy = tf.reduce_sum(tf.cast(correct, tf.float32)) / \
            float(BATCH_SIZE*NUM_POINT)
        tf.summary.scalar('segmentation accuracy', accuracy)

        # Get training operator
        learning_rate = get_learning_rate(batch)
        tf.summary.scalar('learning_rate', learning_rate)

        OPTIMIZER = 'adam'

        if OPTIMIZER == 'momentum':
            optimizer = tf.train.MomentumOptimizer(learning_rate,
                momentum=MOMENTUM)
        elif OPTIMIZER == 'adam':
            optimizer = tf.train.AdamOptimizer(learning_rate)
        train_op = optimizer.minimize(loss, global_step=batch)

#         # Add ops to save and restore all the variables.
#         saver = tf.train.Saver()

#         # Create a session
#         config = tf.ConfigProto()
#         config.gpu_options.allow_growth = True
#         config.allow_soft_placement = True
#         config.log_device_placement = False
#         sess = tf.Session(config=config)

        # Add summary writers
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'train'), sess.graph)
        test_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'test'), sess.graph)

        # Init variables
#         if RESTORE_MODEL_PATH is None:
#             sess.run(init)
#         else:
#             saver.restore(sess, RESTORE_MODEL_PATH)
#             print("Frustum Pointnet sucessfully restored!")

        all_vars = tf.trainable_variables()
        for v in all_vars:
            print("Variable: ", v)
            train_saver = tf.train.Saver([v])
            try:
                train_saver.restore(sess, CHECKPOINT_PATH)
            except:
                try:
                    train_saver.restore(sess, RESTORE_MODEL_PATH)
                except:
                    print("Couldn't restore")
                

    print("RUNTRHOUGH SUCCESSFUL ----------------------------------------")

def main(_):

    params = monodepth_parameters(
        encoder=args.encoder,
        height = INPUT_HEIGHT,
        width = INPUT_WIDTH,
        batch_size=args.batch_size,
        num_threads=args.num_threads,
        num_epochs=args.num_epochs,
        do_stereo=args.do_stereo,
        wrap_mode=args.wrap_mode,
        use_deconv=args.use_deconv,
        alpha_image_loss=args.alpha_image_loss,
        disp_gradient_loss_weight=args.disp_gradient_loss_weight,
        lr_loss_weight=args.lr_loss_weight,
        full_summary=args.full_summary)
    

    both_models_e2e(params)

if __name__ == '__main__':
    tf.app.run()
