from __future__ import absolute_import, division, print_function

# only keep warnings and errors
import os
os.environ['TF_CPP_MIN_LOG_LEVEL']='1'
os.environ["CUDA_VISIBLE_DEVICES"]= "0, 1"

import sys

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(BASE_DIR)
sys.path.append(BASE_DIR)
sys.path.append("/home/EE367_project/frustum-pointnets/models")
sys.path.append("/home/EE367_project/frustum-pointnets/kitti/")
sys.path.append("/home/EE367_project/frustum-pointnets/train/")
sys.path.append("/home/EE367_project/monodepth/")

import numpy as np
import argparse
import re
import time
import tensorflow as tf
import tensorflow.contrib.slim as slim
tf.enable_eager_execution()

from monodepth_model_e2e import *
from monodepth_dataloader_e2e import *

import glob
import importlib
import pandas as pd 
from datetime import datetime
import provider_e2e
from train_util_e2e import get_batch
import time
import pdb

import cv2
from PIL import Image

from data_prep_e2e_tf import depthmap_to_frustum_pointcloud
from data_prep_e2e_tf_rot import depthmap_to_frustum_pointcloud_rot

parser = argparse.ArgumentParser(description='Monodepth TensorFlow implementation.')

parser.add_argument('--mode',                      type=str,   help='train or test', default='test')
parser.add_argument('--model_name',                type=str,   help='model name', default='monodepth_inference_e2e')
parser.add_argument('--encoder',                   type=str,   help='type of encoder, vgg or resnet50', default='vgg')
parser.add_argument('--dataset',                   type=str,   help='dataset to train on, kitti, or cityscapes', default='kitti')
#parser.add_argument('--data_path',                 type=str,   help='path to the data', required=True)
#parser.add_argument('--filenames_file',            type=str,   help='path to the filenames text file', required=True)
#parser.add_argument('--input_height',              type=int,   help='input height', default=256)
#parser.add_argument('--input_width',               type=int,   help='input width', default=512)
parser.add_argument('--batch_size',                type=int,   help='batch size', default=16)
parser.add_argument('--num_epochs',                type=int,   help='number of epochs', default=50)
parser.add_argument('--learning_rate',             type=float, help='initial learning rate', default=1e-4)
parser.add_argument('--lr_loss_weight',            type=float, help='left-right consistency weight', default=1.0)
parser.add_argument('--alpha_image_loss',          type=float, help='weight between SSIM and L1 in the image loss', default=0.85)
parser.add_argument('--disp_gradient_loss_weight', type=float, help='disparity smoothness weigth', default=0.1)
parser.add_argument('--do_stereo',                             help='if set, will train the stereo model', action='store_true')
parser.add_argument('--wrap_mode',                 type=str,   help='bilinear sampler wrap mode, edge or border', default='border')
parser.add_argument('--use_deconv',                            help='if set, will use transposed convolutions', action='store_true')
parser.add_argument('--num_gpus',                  type=int,   help='number of GPUs to use for training', default=2)
parser.add_argument('--num_threads',               type=int,   help='number of threads to use for data loading', default=8)
parser.add_argument('--output_directory',          type=str,   help='output directory for test disparities, if empty outputs to checkpoint folder', default='')
#parser.add_argument('--log_directory',             type=str,   help='directory to save checkpoints and summaries', default='')
#parser.add_argument('--checkpoint_path',           type=str,   help='path to a specific checkpoint to load', default='')
parser.add_argument('--retrain',                               help='if used with checkpoint_path, will restart training from step zero', action='store_true')
parser.add_argument('--full_summary',                          help='if set, will keep more data for each summary. Warning: the file can become very large', action='store_true')

args = parser.parse_args()

# Monodepth parameters
DATA_PATH = "/home/EE367_project/frustum-pointnets/dataset/KITTI/"
FILENAMES_FILE = "monodepth/image_sets/train.txt"
#CHECKPOINT_MONODEPTH = "/home/EE367_project/monodepth/log/model_kitti_test/model-23200"
CHECKPOINT_MONODEPTH = "/home/EE367_project/monodepth/model_kitti"
LOG_DIR = "monodepth/log_e2e/"
INPUT_HEIGHT = 256
INPUT_WIDTH = 1024

# Frustum Parameters
FRUSTUM_MODEL = "frustum_pointnets_v1_e2e"
EPOCH_CNT = 0
BATCH_SIZE = args.batch_size
#BATCH_SIZE = 2
NUM_POINT = 1024
MAX_EPOCH = 50
BASE_LEARNING_RATE = 0.001
GPU_INDEX = 1
MOMENTUM = 0.9
OPTIMIZER = 'adam'
DECAY_STEP = 200000
DECAY_RATE = 0.7
NUM_CHANNEL = 3
NUM_CLASSES = 2 # segmentation has two classes
#CHECKPOINT_FRUSTUM = "/home/EE367_project/frustum-pointnets/snapshots/log_v1/model.ckpt"
CHECKPOINT_FRUSTUM = "/home/EE367_project/frustum-pointnets/log/1552348871.78_epoch_2_model.ckpt"

MODEL = importlib.import_module(FRUSTUM_MODEL) # import network module
MODEL_FILE = os.path.join(ROOT_DIR, 'models', FRUSTUM_MODEL+'.py')
if not os.path.exists(LOG_DIR): os.mkdir(LOG_DIR)
os.system('cp %s %s' % (MODEL_FILE, LOG_DIR)) # bkp of model def
os.system('cp %s %s' % (os.path.join(BASE_DIR, 'train.py'), LOG_DIR))
#LOG_FOUT = open(os.path.join(LOG_DIR, 'log_train.txt'), 'w')
#LOG_FOUT.write(str(FLAGS)+'\n')

BN_INIT_DECAY = 0.5
BN_DECAY_DECAY_RATE = 0.5
BN_DECAY_DECAY_STEP = float(DECAY_STEP)
BN_DECAY_CLIP = 0.99

# Load Frustum Datasets. Use default data paths.

pickle_train_file = "/home/EE367_project/frustum-pointnets/kitti/frustum_preprocessed_data_training.pkl"
pickle_val_file = "/home/EE367_project/frustum-pointnets/kitti/pickles/online_trainin_data_validation.pkl"

ROTATE_TO_CENTER = True


TRAIN_DATASET = provider_e2e.FrustumDataset(split='training', pkl_file_path = pickle_train_file,
    rotate_to_center=ROTATE_TO_CENTER, random_flip=False, random_shift=False, one_hot=True)

TEST_DATASET = provider_e2e.FrustumDataset(split='training', pkl_file_path = pickle_val_file, rotate_to_center=ROTATE_TO_CENTER, one_hot=True)

print(len(TRAIN_DATASET))
print(len(TEST_DATASET))

print("Datasets loaded successfully")

def read_image_np(image_path):

    img = Image.open(image_path)
    img = img.resize([INPUT_WIDTH, INPUT_HEIGHT], Image.ANTIALIAS)
    image = np.array(img).astype(np.float32) / 255.0

    return image

def log_string(out_str):
    #LOG_FOUT.write(out_str+'\n')
    #LOG_FOUT.flush()
    print(out_str)

def get_learning_rate(batch):
    learning_rate = tf.train.exponential_decay(
                        BASE_LEARNING_RATE,  # Base learning rate.
                        batch * BATCH_SIZE,  # Current index into the dataset.
                        DECAY_STEP,          # Decay step.
                        DECAY_RATE,          # Decay rate.
                        staircase=True)
    learing_rate = tf.maximum(learning_rate, 0.00001) # CLIP THE LEARNING RATE!
    return learning_rate        

def get_bn_decay(batch):
    bn_momentum = tf.train.exponential_decay(
                      BN_INIT_DECAY,
                      batch*BATCH_SIZE,
                      BN_DECAY_DECAY_STEP,
                      BN_DECAY_DECAY_RATE,
                      staircase=True)
    bn_decay = tf.minimum(BN_DECAY_CLIP, 1 - bn_momentum)
    return bn_decay

def post_process_disparity(disp):
    _, h, w = disp.shape
    l_disp = disp[0,:,:]
    r_disp = np.fliplr(disp[1,:,:])
    m_disp = 0.5 * (l_disp + r_disp)
    l, _ = np.meshgrid(np.linspace(0, 1, w), np.linspace(0, 1, h))
    l_mask = 1.0 - np.clip(20 * (l - 0.05), 0, 1)
    r_mask = np.fliplr(l_mask)
    return r_mask * l_disp + l_mask * r_disp + (1.0 - l_mask - r_mask) * m_disp

def count_text_lines(file_path):
    f = open(file_path, 'r')
    lines = f.readlines()
    f.close()
    return len(lines)

def train(params):
    """Test function."""
      
    # load and restore the monodepth graph
    tf_monodepth_graph = tf.Graph()
    with tf_monodepth_graph.as_default():
        
        #with tf.variable_scope("monodepth"):


        #dataloader = MonodepthDataloader(DATA_PATH, FILENAMES_FILE, params, args.dataset, args.mode)
        #left  = dataloader.left_image_batch
        right_img_pl = tf.placeholder(tf.float32, shape=(BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))
        left_img_pl = tf.placeholder(tf.float32, shape=(BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))

        model = MonodepthModel(params, args.mode, left_img_pl, None)

        # get real depth estimation
        real_depth = tf.identity(model.real_depth_est, name = "real_depth")
        
        # SESSION
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        config.allow_soft_placement = True
        config.log_device_placement = False
        sess = tf.Session(config=config)


        # INIT
        sess.run(tf.global_variables_initializer())
        sess.run(tf.local_variables_initializer())
        coordinator = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coordinator)


        real_depth = tf.reshape(real_depth, [BATCH_SIZE, 375, 1242])

        #with tf.variable_scope("frustum"):

        npy_depth_pl, rgb_img_pl, box2d_pl, box3d_pl, rot_angle_pl, calib_pl, one_hot_vec_pl, centers_pl, \
        heading_class_label_pl, heading_residual_label_pl, \
        size_class_label_pl, size_residual_label_pl = \
            MODEL.placeholder_inputs(BATCH_SIZE, NUM_POINT)

        is_training_pl = tf.placeholder(tf.bool, shape = ())        # Note the global_step=batch parameter to minimize. 
        # That tells the optimizer to increment the 'batch' parameter
        # for you every time it trains.
        batch = tf.get_variable('batch', [],
            initializer=tf.constant_initializer(0), trainable=False)
        bn_decay = get_bn_decay(batch)

        tf.summary.scalar('bn_decay', bn_decay)

        # Get model and losses (note: no placehoder anymore as first argument!)
        if ROTATE_TO_CENTER:
            pointcloud, labels = \
                depthmap_to_frustum_pointcloud_rot(real_depth, box2d_pl, box3d_pl, \
                                                   calib_pl, rot_angle_pl, npoints = NUM_POINT)
        else:
            pointcloud, labels = \
                depthmap_to_frustum_pointcloud(real_depth, box2d_pl, box3d_pl, \
                                               calib_pl, npoints = NUM_POINT)


        end_points = MODEL.get_model(pointcloud, one_hot_vec_pl,
            is_training_pl, bn_decay=bn_decay)
        loss = MODEL.get_loss(labels, centers_pl,
            heading_class_label_pl, heading_residual_label_pl,
            size_class_label_pl, size_residual_label_pl, end_points)
        tf.summary.scalar('loss', loss)

        variables = tf.trainable_variables() 
        lossL2 = tf.add_n([ tf.nn.l2_loss(v) for v in variables ]) * 0.001
        loss += lossL2

        losses = tf.get_collection('losses')
        total_loss = tf.add_n(losses, name='total_loss')
        tf.summary.scalar('total_loss', total_loss)

        # Write summaries of bounding box IoU and segmentation accuracies
        iou2ds, iou3ds = tf.py_func(provider_e2e.compute_box3d_iou, [\
            end_points['center'], \
            end_points['heading_scores'], end_points['heading_residuals'], \
            end_points['size_scores'], end_points['size_residuals'], \
            centers_pl, \
            heading_class_label_pl, heading_residual_label_pl, \
            size_class_label_pl, size_residual_label_pl], \
            [tf.float32, tf.float32])
        end_points['iou2ds'] = iou2ds 
        end_points['iou3ds'] = iou3ds 
        tf.summary.scalar('iou_2d', tf.reduce_mean(iou2ds))
        tf.summary.scalar('iou_3d', tf.reduce_mean(iou3ds))

        correct = tf.equal(tf.argmax(end_points['mask_logits'], 2),
            tf.to_int64(labels))
        accuracy = tf.reduce_sum(tf.cast(correct, tf.float32)) / \
            float(BATCH_SIZE*NUM_POINT)
        tf.summary.scalar('segmentation accuracy', accuracy)

        # Get training operator
        learning_rate = get_learning_rate(batch)
        tf.summary.scalar('learning_rate', learning_rate)

        OPTIMIZER = 'adam'

        if OPTIMIZER == 'momentum':
            optimizer = tf.train.MomentumOptimizer(learning_rate,
                momentum=MOMENTUM)
        elif OPTIMIZER == 'adam':
            optimizer = tf.train.AdamOptimizer(learning_rate)
        train_op = optimizer.minimize(loss, global_step=batch)


        # Add summary writers
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'train'), sess.graph)
        test_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'test'), sess.graph)

#         all_vars_mono = tf.global_variables(scope='monodepth')
#         all_vars_frustum = tf.global_variables(scope='frustum')
        
        all_vars = tf.global_variables()
#         var_name = [all_vars[i].name for i in range(len(all_vars))]
#         var_name_unique = set(var_name)
        
#         ambiguous_vars = []
        
#         for v in all_vars:
#             num_not_found = 0
#             print("Variable: ", v)
#             train_saver = tf.train.Saver([v])
#             try:
#                 if v.name.find("beta1_power") == -1 and v.name.find("beta2_power") == -1:
#                     train_saver.restore(sess, CHECKPOINT_MONODEPTH)
#                     print("Restored")
#                 else:
#                     print("Found weird variable")
#             except:
#                 num_not_found = num_not_found + 1
#             try:
#                 train_saver.restore(sess, CHECKPOINT_FRUSTUM)
#                 print("Restored")
#             except:
#                 num_not_found = num_not_found + 1
                
#             if (num_not_found == 0):
#                 print("STRANGE!!!!!")
#                 ambiguous_vars.append(v)


#         for v in all_vars_mono:
#             v = tf.Variable(v, name=v.name[10:])
#             print("Variable: ", v)
#             train_saver = tf.train.Saver([v])
#             train_saver.restore(sess, CHECKPOINT_MONODEPTH)
#         for v in all_vars_frustum:
#             while v.name[:8] = "frustum/":
#                 v = tf.Variable(v, name=v.name[8:])
#             print("Variable: ", v)
#             train_saver = tf.train.Saver([v])
#             train_saver.restore(sess, CHECKPOINT_FRUSTUM)
                    
        train_saver = tf.train.Saver()
        model_name = "baseline_model_e2e_original.ckpt"
        train_saver.restore(sess, os.path.join(LOG_DIR, model_name))
        print("Model sucessfully restored")
        #tf.global_variables_initializer()
        
        #load_path = "monodepth/log_e2e/baseline_model_e2e.ckpt"
#         save_path = train_saver.save(sess, os.path.join(LOG_DIR, model_name))
#         log_string("Model saved in file: %s" % save_path)

#         train_saver.restore(sess, os.path.join(LOG_DIR, model_name))
#         print("Model sucessfully restored")
        #batch = tf.get_variable('batch', [], initializer=tf.constant_initializer(0), trainable=False, reuse = True)
        #bn_decay = get_bn_decay(batch)
                    
        # Todo: #add np arrays, calibs to feed_dict
        ops = {'right_img_pl': right_img_pl,
               'left_img_pl': left_img_pl,
               'npy_depth_pl': npy_depth_pl,
               'calib_pl': calib_pl, 
               'box2d_pl' : box2d_pl,
               'box3d_pl' : box3d_pl,
               'rot_angle_pl': rot_angle_pl,
               'one_hot_vec_pl': one_hot_vec_pl,
               'centers_pl': centers_pl,
               'heading_class_label_pl': heading_class_label_pl,
               'heading_residual_label_pl': heading_residual_label_pl,
               'size_class_label_pl': size_class_label_pl,
               'size_residual_label_pl': size_residual_label_pl,
               'is_training_pl': is_training_pl,
               'logits': end_points['mask_logits'],
               'centers_pred': end_points['center'],
               'loss': loss,
               'train_op': train_op,
               'merged': merged,
               'labels_output' : labels,
               'step': batch,
               'end_points': end_points, 
               'real_depth': real_depth, 
               'disp1': model.disp1, 
               'frustum_pc': pointcloud}
        
        print("Running evaluation once before training")
        eval_one_epoch(sess, ops, test_writer)
        
        pdb.set_trace()
        
        for epoch in range(MAX_EPOCH):
            model_ts = time.time()
            log_string('**** EPOCH %03d ****' % (epoch))
            sys.stdout.flush()
             
            train_one_epoch(sess, ops, train_writer)
            #eval_one_epoch(sess, ops, test_writer)

            # Save the variables to disk.
            if epoch >= 2:
                saver = tf.train.Saver()
                model_name = str(model_ts) + "_epoch_" + str(epoch) + "_model.ckpt" 
                save_path = saver.save(sess, os.path.join(LOG_DIR, model_name))
                log_string("Model saved in file: %s" % save_path)

def train_one_epoch(sess, ops, train_writer):
    ''' Training for one epoch on the frustum dataset.
    ops is dict mapping from string to tf ops
    '''
    is_training = True
    log_string(str(datetime.now()))
    
    # Shuffle train samples
    train_idxs = np.arange(0, len(TRAIN_DATASET))
    #np.random.shuffle(train_idxs)
    num_batches = len(TRAIN_DATASET)/BATCH_SIZE

    # To collect statistics
    total_correct = 0
    total_seen = 0
    loss_sum = 0
    iou2ds_sum = 0
    iou3ds_sum = 0
    iou3d_correct_cnt = 0

    # Training with batches
    print("Training epoch")
    for batch_idx in range(int(num_batches)):
        start_idx = batch_idx * BATCH_SIZE
        end_idx = (batch_idx+1) * BATCH_SIZE

        npy_files, _, data_idx_arr, calibs, box2d, box3d, batch_center, \
        batch_hclass, batch_hres, \
        batch_sclass, batch_sres, \
        batch_rot_angle, batch_one_hot_vec = \
            get_batch(TRAIN_DATASET, train_idxs, start_idx, end_idx, NUM_CHANNEL)
        
        right_img_array = np.zeros((BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))
        left_img_array = np.zeros((BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))
        
        for i in range(BATCH_SIZE):
        #for i in range(1):
            line = "%06d" % (data_idx_arr[i])
            left_image_path  = DATA_PATH + "/object/training/image_2/" + line + ".png"
            right_image_path = DATA_PATH + "/object/training/image_3/" + line + ".png"
            left_img  = read_image_np(left_image_path)
            right_img = read_image_np(right_image_path)
            right_img_array[i,...] = right_img
            left_img_array[i,...] = left_img
            #right_img_array[1,...] = np.fliplr(right_img)
            #left_img_array[1,...] = np.fliplr(left_img)
            
        
        feed_dict = {ops['right_img_pl']: right_img_array, 
                     ops['left_img_pl']: left_img_array, 
                     ops['npy_depth_pl']: npy_files,
                     ops['calib_pl']: calibs, 
                     ops['box2d_pl']: box2d,
                     ops['box3d_pl']: box3d,
                     ops['rot_angle_pl']: batch_rot_angle,
                     ops['one_hot_vec_pl']: batch_one_hot_vec,
                     ops['centers_pl']: batch_center,
                     ops['heading_class_label_pl']: batch_hclass,
                     ops['heading_residual_label_pl']: batch_hres,
                     ops['size_class_label_pl']: batch_sclass,
                     ops['size_residual_label_pl']: batch_sres,
                     ops['is_training_pl']: is_training}
        
        t1 = time.time()
        pointcloud, real_depth_debug, disp1_debug, summary, labels_output, step, _, loss_val, logits_val, centers_pred_val, \
        iou2ds, iou3ds = \
            sess.run([ops['frustum_pc'], ops['real_depth'], ops['disp1'], ops['merged'], ops['labels_output'], 
                ops['step'], ops['train_op'], ops['loss'],
                ops['logits'], ops['centers_pred'],
                ops['end_points']['iou2ds'], ops['end_points']['iou3ds']], 
                feed_dict=feed_dict)
        
        # DEBUG
#         np.save("debug/real_depth_%d.npy" % (batch_idx), real_depth_debug)
#         np.save("debug/disp1_%d.npy" % (batch_idx), disp1_debug)
#         np.save("debug/data_idx_%d.npy" % (batch_idx), data_idx_arr)
#         np.save("debug/pc_%d.npy" % (batch_idx), pointcloud)
#         np.save("debug/npy_files_%d.npy" % (batch_idx), npy_files)
        
        
        t2 = time.time()
        tdelta = t2 - t1
        print("Elapsed time: ", tdelta)
        train_writer.add_summary(summary, step)

        preds_val = np.argmax(logits_val, 2)
        correct = np.sum(preds_val == labels_output)
        total_correct += correct
        total_seen += (BATCH_SIZE*NUM_POINT)
        loss_sum += loss_val
        iou2ds_sum += np.sum(iou2ds)
        iou3ds_sum += np.sum(iou3ds)
        iou3d_correct_cnt += np.sum(iou3ds>=0.7)

        if (batch_idx+1)%10 == 0:
            log_string(' -- %03d / %03d --' % (batch_idx+1, num_batches))
            log_string('mean loss: %f' % (loss_sum / 10))
            log_string('segmentation accuracy: %f' % \
                (total_correct / float(total_seen)))
            log_string('box IoU (ground/3D): %f / %f' % \
                (iou2ds_sum / float(BATCH_SIZE*10), iou3ds_sum / float(BATCH_SIZE*10)))
            log_string('box estimation accuracy (IoU=0.7): %f' % \
                (float(iou3d_correct_cnt)/float(BATCH_SIZE*10)))
            total_correct = 0
            total_seen = 0
            loss_sum = 0
            iou2ds_sum = 0
            iou3ds_sum = 0
            iou3d_correct_cnt = 0
        
        
def eval_one_epoch(sess, ops, test_writer):
    ''' Simple evaluation for one epoch on the frustum dataset.
    ops is dict mapping from string to tf ops """
    '''
    global EPOCH_CNT
    is_training = False
    log_string(str(datetime.now()))
    log_string('---- EPOCH %03d EVALUATION ----'%(EPOCH_CNT))
    test_idxs = np.arange(0, len(TEST_DATASET))
    num_batches = len(TEST_DATASET)/BATCH_SIZE

    # To collect statistics
    total_correct = 0
    total_seen = 0
    loss_sum = 0
    total_seen_class = [0 for _ in range(NUM_CLASSES)]
    total_correct_class = [0 for _ in range(NUM_CLASSES)]
    iou2ds_sum = 0
    iou3ds_sum = 0
    iou3d_correct_cnt = 0
    
    # Simple evaluation with batches 
    for batch_idx in range(int(num_batches)):
    #for batch_idx in range(20):
        
        log_string('---- BATCH %03d of %03d EVALUATION ----'%(batch_idx, num_batches))
        
        start_idx = batch_idx * BATCH_SIZE
        end_idx = (batch_idx+1) * BATCH_SIZE

        npy_arrays, imgs, data_idx_arr, calibs, box2d, box3d, batch_center, \
        batch_hclass, batch_hres, \
        batch_sclass, batch_sres, \
        batch_rot_angle, batch_one_hot_vec = \
            get_batch(TEST_DATASET, test_idxs, start_idx, end_idx, NUM_CHANNEL)
        
        right_img_array = np.zeros((BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))
        left_img_array = np.zeros((BATCH_SIZE, INPUT_HEIGHT, INPUT_WIDTH, 3))
        
        for i in range(BATCH_SIZE):
            line = str(int(data_idx_arr[i])).zfill(6)
            left_image_path  = DATA_PATH + "/object/training/image_2/" + line + ".png"
            right_image_path = DATA_PATH + "/object/training/image_3/" + line + ".png"
            left_img  = read_image_np(left_image_path)
            right_img = read_image_np(right_image_path)
            right_img_array[i,...] = right_img
            left_img_array[i,...] = left_img

        feed_dict = {ops['right_img_pl']: right_img_array, 
                     ops['left_img_pl']: left_img_array, 
                     ops['calib_pl']: calibs, 
                     ops['box2d_pl']: box2d,
                     ops['box3d_pl']: box3d,
                     ops['rot_angle_pl']: batch_rot_angle,
                     ops['one_hot_vec_pl']: batch_one_hot_vec,
                     ops['centers_pl']: batch_center,
                     ops['heading_class_label_pl']: batch_hclass,
                     ops['heading_residual_label_pl']: batch_hres,
                     ops['size_class_label_pl']: batch_sclass,
                     ops['size_residual_label_pl']: batch_sres,
                     ops['is_training_pl']: is_training}

        summary, labels_output, step, loss_val, logits_val, iou2ds, iou3ds = \
            sess.run([ops['merged'], ops['labels_output'], ops['step'],
                ops['loss'], ops['logits'], 
                ops['end_points']['iou2ds'], ops['end_points']['iou3ds']],
                feed_dict=feed_dict)
        test_writer.add_summary(summary, step)

        preds_val = np.argmax(logits_val, 2)
        correct = np.sum(preds_val == labels_output)
        total_correct += correct
        total_seen += (BATCH_SIZE*NUM_POINT)
        loss_sum += loss_val
        ## replace batch_label with labels_output
        for l in range(NUM_CLASSES):
            total_seen_class[l] += np.sum(labels_output==l)
            total_correct_class[l] += (np.sum((preds_val==l) & (labels_output==l)))
        iou2ds_sum += np.sum(iou2ds)
        iou3ds_sum += np.sum(iou3ds)
        iou3d_correct_cnt += np.sum(iou3ds>=0.7)

        for i in range(BATCH_SIZE):
            segp = preds_val[i,:]
            segl = labels_output[i,:] 
            part_ious = [0.0 for _ in range(NUM_CLASSES)]
            for l in range(NUM_CLASSES):
                if (np.sum(segl==l) == 0) and (np.sum(segp==l) == 0): 
                    part_ious[l] = 1.0 # class not present
                else:
                    part_ious[l] = np.sum((segl==l) & (segp==l)) / \
                        float(np.sum((segl==l) | (segp==l)))

    log_string('eval mean loss: %f' % (loss_sum / float(num_batches)))
    log_string('eval segmentation accuracy: %f'% \
        (total_correct / float(total_seen)))
    log_string('eval segmentation avg class acc: %f' % \
        (np.mean(np.array(total_correct_class) / \
            np.array(total_seen_class,dtype=np.float))))
    log_string('eval box IoU (ground/3D): %f / %f' % \
        (iou2ds_sum / float(num_batches*BATCH_SIZE), iou3ds_sum / \
            float(num_batches*BATCH_SIZE)))
    log_string('eval box estimation accuracy (IoU=0.7): %f' % \
        (float(iou3d_correct_cnt)/float(num_batches*BATCH_SIZE)))
         
    EPOCH_CNT += 1


if __name__ == "__main__":
    #log_string('pid: %s'%(str(os.getpid())))
    params = monodepth_parameters(
        encoder=args.encoder,
        height = INPUT_HEIGHT,
        width = INPUT_WIDTH,
        batch_size=args.batch_size,
        num_threads=args.num_threads,
        num_epochs=args.num_epochs,
        do_stereo=args.do_stereo,
        wrap_mode=args.wrap_mode,
        use_deconv=args.use_deconv,
        alpha_image_loss=args.alpha_image_loss,
        disp_gradient_loss_weight=args.disp_gradient_loss_weight,
        lr_loss_weight=args.lr_loss_weight,
        full_summary=args.full_summary)

    train(params)
